@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2023, University of Amsterdam
+    Copyright (c)  1985-2024, University of Amsterdam
 			      VU University Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
@@ -533,6 +533,56 @@ ssu_or_det_failed(DECL_LD LocalFrame fr)
 		*         FOREIGN CALLS         *
 		*********************************/
 
+#define vmi_fopen(fr, def) LDFUNC(vmi_fopen, fr, def)
+
+static void
+vmi_fopen(DECL_LD LocalFrame fr, Definition def)
+{ FliFrame ffr;
+
+#ifdef O_DEBUGGER
+  if ( debugstatus.debugging )
+  { lTop = (LocalFrame)argFrameP(fr, def->functor->arity);
+    BFR = newChoice(CHP_DEBUG, fr);
+    ffr = (FliFrame)lTop;
+  } else
+#endif
+  { ffr = (FliFrame)argFrameP(fr, def->functor->arity);
+  }
+
+#if O_DEBUG
+  if ( exception_term )
+  { Sdprintf("Exception at entry of %s\n",  predicateName(def));
+    PL_write_term(Serror, exception_term, 1200, PL_WRT_NEWLINE);
+  }
+#endif
+
+  DEBUG(CHK_SECURE, assert(def->functor->arity < 100));
+
+  lTop = (LocalFrame)(ffr+1);
+  ffr->size = 0;
+  NoMark(ffr->mark);
+  ffr->parent = fli_context;
+  FLI_SET_VALID(ffr);
+  fli_context = ffr;
+}
+
+
+static void
+error_foreign_return_code(intptr_t rc)
+{ fid_t fid = PL_open_foreign_frame();
+
+  if ( (fid = PL_open_foreign_frame()) )
+  { term_t ex;
+    int rc2 = ( (ex=PL_new_term_ref()) &&
+		PL_put_intptr(ex, rc) &&
+		PL_error(NULL, 0, NULL, ERR_DOMAIN,
+			 ATOM_foreign_return_value, ex) );
+    (void)rc2;
+    PL_close_foreign_frame(fid);
+  }
+}
+
+
 #define CALL_FCUTTED(argc, f, c) \
   { switch(argc) \
     { case 0: \
@@ -2146,7 +2196,7 @@ exception_hook(DECL_LD qid_t pqid, term_t fr, term_t catchfr_ref)
 	PL_cut_query(qid);
       }
 
-      if ( ex && !PL_same_term(ex, exception_term) )
+      if ( ex && (!exception_term || !PL_same_term(ex, exception_term)) )
       {	PL_raise_exception(ex);	/* copy term again */
 	wstate.flags |= WAKEUP_STATE_SKIP_EXCEPTION;
 	rc = TRUE;			/* handled */
@@ -2985,6 +3035,9 @@ typedef struct register_file
   jmp_buf    exit_vm_buf;		/* jump target for exiting PL_next_solution */
 # define     EXIT_VM_BUF		(REGISTERS.exit_vm_buf)
 #endif
+  /* Not initialized below here */
+  size_t     pl_ar_buf[GMP_STACK_ALLOC];
+#define __PL_ar_buf	(REGISTERS.pl_ar_buf)
 } register_file;
 
 
@@ -3176,13 +3229,15 @@ API_STUB(int)
 
 int
 PL_next_solution(DECL_LD qid_t qid)
-{ register_file REGISTERS =
-  { .qid = qid,
-    .fndet_context.engine = LD
-#if VMCODE_IS_ADDRESS
-    , .nop1 = 0, .nop2 = 0
+{ register_file REGISTERS;
+  memset(&REGISTERS, 0, offsetof(register_file,pl_ar_buf));
+  QID = qid;
+  FNDET_CONTEXT.engine = LD;
+#ifdef VMCODE_IS_ADDRESS
+  REGISTERS.nop1 = 0;
+  REGISTERS.nop2 = 0;
 #endif
-  };					/* Active registers */
+
   Code PC;				/* program counter */
   exception_frame THROW_ENV;		/* PL_thow() environment */
 
